"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const InputWatcher_1 = __importDefault(require("../browser/InputWatcher"));
const ResizeWatcher_1 = __importDefault(require("../browser/ResizeWatcher"));
const engine_1 = require("../engine");
const astTypes_1 = require("../parser/astTypes");
const parser_1 = __importDefault(require("../parser/parser"));
const table_1 = __importDefault(require("../ui/table"));
const util_1 = require("../util");
class SubTableEngine {
    constructor(table, optionalHandler) {
        this.table = table;
        this.inputWatcher = new InputWatcher_1.default(table);
        this.timer = 0;
        this.engine = null;
        // wait until user is no longer pressing anything before
        // showing the alert().
        // https://bugzilla.mozilla.org/show_bug.cgi?id=1346228
        const messageHandler = {
            onWin() {
                alert('You won the game! Congratulations!');
            },
            onMessage: (msg) => __awaiter(this, void 0, void 0, function* () {
                return util_1.pollingPromise(10, () => {
                    if (this.inputWatcher.isSomethingPressed()) {
                        return false;
                    }
                    alert(msg);
                    table.focus();
                    return true;
                });
            })
        };
        const handler = optionalHandler ? new util_1.EmptyGameEngineHandler([optionalHandler, messageHandler]) : new util_1.EmptyGameEngineHandler([messageHandler]);
        this.tableUI = new table_1.default(table, handler);
    }
    setGame(code, levelNum) {
        const { data } = parser_1.default.parse(code);
        this.engine = new engine_1.GameEngine(data, this.tableUI);
        this.tableUI.setGameData(data);
        this.engine.setLevel(levelNum);
        if (data.metadata.keyRepeatInterval) {
            this.inputWatcher.setKeyRepeatInterval(data.metadata.keyRepeatInterval);
        }
    }
    start() {
        if (this.timer) {
            this.pause();
        }
        this.table.setAttribute('data-ps-state', 'running');
        const runLoop = () => __awaiter(this, void 0, void 0, function* () {
            const pendingKey = this.inputWatcher.pollControls();
            if (pendingKey) {
                this.engine && this.engine.press(pendingKey);
            }
            yield this.getEngine().tick();
            this.timer = window.requestAnimationFrame(runLoop);
        });
        this.timer = window.requestAnimationFrame(runLoop);
    }
    pause() {
        this.table.setAttribute('data-ps-state', 'paused');
        cancelAnimationFrame(this.timer);
        this.timer = 0;
    }
    getEngine() {
        if (!this.engine) {
            throw new Error(`BUG: Engine has not been created yet`);
        }
        return this.engine;
    }
    dispose() {
        this.pause();
        this.inputWatcher.dispose();
    }
}
class SyncTableEngine {
    constructor(table, handler) {
        this.table = table;
        this.subEngine = new SubTableEngine(table, handler);
        this.resizeWatcher = new ResizeWatcher_1.default(table, this.handleResize.bind(this));
        this.handler = new util_1.EmptyGameEngineHandler(handler ? [handler, this.subEngine.tableUI] : [this.subEngine.tableUI]);
        this.boundPause = this.pause.bind(this);
        this.boundResume = this.resume.bind(this);
        this.table.addEventListener('blur', this.boundPause);
        this.table.addEventListener('focus', this.boundResume);
    }
    setGame(source, level = 0) {
        this.subEngine.setGame(source, level);
        const engine = this.subEngine.getEngine();
        if (engine.getCurrentLevel().type === astTypes_1.LEVEL_TYPE.MAP) {
            const currentLevel = engine.getCurrentLevelCells();
            this.resizeWatcher.setLevel(currentLevel.length, currentLevel[0].length);
        }
        this.subEngine.start();
        this.table.focus();
        this.resizeWatcher.trigger();
    }
    dispose() {
        this.subEngine.dispose();
        this.resizeWatcher.dispose();
        this.table.removeEventListener('blur', this.boundPause);
        this.table.removeEventListener('focus', this.boundResume);
    }
    pause() {
        this.subEngine.pause();
        this.handler.onPause();
    }
    resume() {
        this.subEngine.start();
        this.handler.onResume();
    }
    handleResize(width) {
        if (!this.subEngine.getEngine().isCurrentLevelAMessage()) {
            this.table.setAttribute('style', `width: ${width}px;`);
        }
    }
}
exports.default = SyncTableEngine;
//# sourceMappingURL=SyncTableEngine.js.map