"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const rule_1 = require("../models/rule");
const astTypes_1 = require("../parser/astTypes");
const serializer_1 = __importDefault(require("../parser/serializer"));
const table_1 = __importDefault(require("../ui/table"));
const util_1 = require("../util");
const InputWatcher_1 = __importDefault(require("./InputWatcher"));
const ResizeWatcher_1 = __importDefault(require("./ResizeWatcher"));
class ProxyCellish {
    constructor(rowIndex, colIndex, sprites) {
        this.rowIndex = rowIndex;
        this.colIndex = colIndex;
        this.sprites = sprites;
    }
    getSprites() {
        return this.sprites;
    }
    getSpritesAsSet() {
        return new Set(this.sprites);
    }
    getWantsToMove() {
        return util_1.RULE_DIRECTION.STATIONARY;
    }
}
class WebworkerTableEngine {
    constructor(worker, table, handler) {
        this.worker = worker;
        this.table = table;
        const defaultOnMessage = (msg) => {
            return util_1.pollingPromise(10, () => {
                // Using a pollingPromise is only necessary for alert() because of the keyUp event not firing
                if (this.inputWatcher.isSomethingPressed()) {
                    return false;
                }
                alert(msg);
                return true;
            });
        };
        if (!handler || !handler.onMessage) {
            handler = Object.assign({}, handler, { onMessage: defaultOnMessage });
        }
        // cache
        this.cellCache = [];
        this.levelNum = -123456;
        this.gameData = null;
        this.ui = new table_1.default(table, handler);
        this.resizeWatcher = new ResizeWatcher_1.default(table, this.handleResize.bind(this));
        this.inputWatcher = new InputWatcher_1.default(table);
        this.boundPause = this.pause.bind(this);
        this.boundResume = this.resume.bind(this);
        this.boundMessageListener = this.messageListener.bind(this);
        table.addEventListener('blur', this.boundPause);
        table.addEventListener('focus', this.boundResume);
        worker.addEventListener('message', this.boundMessageListener);
        this.inputInterval = window.setInterval(() => {
            const button = this.inputWatcher.pollControls();
            if (button) {
                this.press(button);
            }
        }, 10);
    }
    setGame(code, level) {
        this.worker.postMessage({ type: util_1.MESSAGE_TYPE.LOAD_GAME, code, level });
    }
    dispose() {
        this.inputWatcher.dispose();
        this.resizeWatcher.dispose();
        clearInterval(this.inputInterval);
    }
    press(button) {
        this.worker.postMessage({ type: util_1.MESSAGE_TYPE.PRESS, button });
    }
    pause() {
        this.worker.postMessage({ type: util_1.MESSAGE_TYPE.PAUSE });
    }
    resume() {
        this.worker.postMessage({ type: util_1.MESSAGE_TYPE.RESUME });
    }
    messageListener({ data }) {
        return __awaiter(this, void 0, void 0, function* () {
            switch (data.type) {
                case util_1.MESSAGE_TYPE.LOAD_GAME:
                    const gameData = serializer_1.default.fromJson(data.payload, '**source not included because of laziness**');
                    this.gameData = gameData;
                    this.ui.setGameData(gameData);
                    break;
                case util_1.MESSAGE_TYPE.ON_LEVEL_CHANGE:
                    this.cellCache = []; // clear the cache since the level dimensions are different
                    this.levelNum = data.level;
                    if (data.cells) {
                        this.ui.onLevelChange(data.level, data.cells.map((row) => row.map((x) => this.convertToCellish(x))), null);
                    }
                    else {
                        this.ui.onLevelChange(data.level, null, data.message);
                    }
                    this.resizeWatcher.trigger(); // force resize
                    this.table.focus();
                    break;
                case util_1.MESSAGE_TYPE.ON_MESSAGE:
                    yield this.ui.onMessage(data.message);
                    this.worker.postMessage({ type: util_1.MESSAGE_TYPE.ON_MESSAGE_DONE });
                    this.table.focus();
                    break;
                case util_1.MESSAGE_TYPE.ON_PRESS:
                    this.ui.onPress(data.direction);
                    break;
                case util_1.MESSAGE_TYPE.ON_TICK:
                    this.ui.onTick(new Set(data.changedCells.map((x) => this.convertToCellish(x))), data.hasAgain, this.convertToA11yMessages(data.a11yMessages));
                    break;
                case util_1.MESSAGE_TYPE.ON_SOUND:
                    yield this.ui.onSound({ soundCode: data.soundCode });
                    break;
                case util_1.MESSAGE_TYPE.ON_PAUSE:
                    this.ui.onPause();
                    break;
                case util_1.MESSAGE_TYPE.ON_RESUME:
                    this.ui.onResume();
                    break;
                case util_1.MESSAGE_TYPE.TICK:
                case util_1.MESSAGE_TYPE.PRESS:
                case util_1.MESSAGE_TYPE.PAUSE:
                case util_1.MESSAGE_TYPE.RESUME:
                    break; // less console noise
                default:
                    console.log(`BUG: Unhandled Event occurred. Ignoring "${data.type}"`, data); // tslint:disable-line:no-console
            }
        });
    }
    convertToA11yMessages(a11yMessages) {
        return a11yMessages.map((message) => {
            switch (message.type) {
                case rule_1.A11Y_MESSAGE_TYPE.ADD:
                    return Object.assign({}, message, { cell: this.convertToCellish(message.cell), sprites: [...message.sprites].map((n) => this.lookupSprite(n)) });
                case rule_1.A11Y_MESSAGE_TYPE.MOVE:
                    return Object.assign({}, message, { oldCell: this.convertToCellish(message.oldCell), newCell: this.convertToCellish(message.newCell), sprite: this.lookupSprite(message.sprite) });
                case rule_1.A11Y_MESSAGE_TYPE.REMOVE:
                    return Object.assign({}, message, { cell: this.convertToCellish(message.cell), sprites: [...message.sprites].map((n) => this.lookupSprite(n)) });
                case rule_1.A11Y_MESSAGE_TYPE.REPLACE:
                    return Object.assign({}, message, { cell: this.convertToCellish(message.cell), replacements: [...message.replacements].map(({ oldSprite, newSprite }) => ({ oldSprite: this.lookupSprite(oldSprite), newSprite: this.lookupSprite(newSprite) })) });
            }
        });
    }
    lookupSprite(name) {
        const sprite = this.getGameData()._getSpriteByName(name);
        if (!sprite) {
            throw new Error(`Could not find sprite "${name}"`);
        }
        return sprite;
    }
    convertToCellish(c) {
        const { rowIndex, colIndex, spriteNames } = c;
        // Lazy-initialize all the proxy cells
        if (!this.cellCache[rowIndex]) {
            this.cellCache[rowIndex] = [];
        }
        if (!this.cellCache[rowIndex][colIndex]) {
            this.cellCache[rowIndex][colIndex] = new ProxyCellish(rowIndex, colIndex, []);
        }
        const cell = this.cellCache[rowIndex][colIndex];
        cell.sprites = spriteNames.map((name) => this.lookupSprite(name));
        return cell;
    }
    getGameData() {
        if (!this.gameData) {
            throw new Error(`BUG: Game Data has not been retreived from the worker yet`);
        }
        return this.gameData;
    }
    isCurrentLevelAMessage() {
        return this.getGameData().levels[this.levelNum].type === astTypes_1.LEVEL_TYPE.MESSAGE;
    }
    handleResize(width) {
        if (!this.isCurrentLevelAMessage()) {
            this.table.setAttribute('style', `width: ${width}px;`);
        }
    }
}
exports.default = WebworkerTableEngine;
//# sourceMappingURL=WebworkerTableEngine.js.map