"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const contro_1 = require("contro");
const util_1 = require("../util");
class InputWatcher {
    constructor(table) {
        const keyboard = new contro_1.Keyboard();
        this.gamepad = new contro_1.Gamepad();
        this.table = table;
        this.polledInput = null;
        this.repeatIntervalInSeconds = null;
        this.possibleKeys = [];
        const keyboardKey = (key) => {
            this.possibleKeys.push(key);
            return keyboard.key(key);
        };
        this.controls = {
            up: { lastPressed: null, button: contro_1.or(this.gamepad.button(contro_1.BUTTON_TYPE.ARROW_UP), keyboardKey('w'), keyboardKey('ArrowUp')) },
            down: { lastPressed: null, button: contro_1.or(this.gamepad.button(contro_1.BUTTON_TYPE.ARROW_DOWN), keyboardKey('s'), keyboardKey('ArrowDown')) },
            left: { lastPressed: null, button: contro_1.or(this.gamepad.button(contro_1.BUTTON_TYPE.ARROW_LEFT), keyboardKey('a'), keyboardKey('ArrowLeft')) },
            right: { lastPressed: null, button: contro_1.or(this.gamepad.button(contro_1.BUTTON_TYPE.ARROW_RIGHT), keyboardKey('d'), keyboardKey('ArrowRight')) },
            action: { lastPressed: null, button: contro_1.or(this.gamepad.button(contro_1.BUTTON_TYPE.CLUSTER_BOTTOM), keyboardKey('x'), keyboardKey(' '), keyboardKey('Enter')) },
            undo: { lastPressed: null, button: contro_1.or(this.gamepad.button(contro_1.BUTTON_TYPE.CLUSTER_LEFT), this.gamepad.button(contro_1.BUTTON_TYPE.BUMPER_TOP_LEFT), keyboardKey('z'), keyboardKey('u')) },
            restart: { lastPressed: null, button: contro_1.or(this.gamepad.button(contro_1.BUTTON_TYPE.CLUSTER_TOP), keyboardKey('r')) }
        };
        this.stickControls = {
            left: { lastPressed: null, stick: this.gamepad.stick(contro_1.STICK_TYPE.LEFT) },
            right: { lastPressed: null, stick: this.gamepad.stick(contro_1.STICK_TYPE.RIGHT) }
        };
        const makeChecker = (control, gameAction) => {
            return () => {
                const now = Date.now();
                const interval = (this.repeatIntervalInSeconds || .25) * 1000;
                const { button, lastPressed } = this.controls[control];
                if (button.query()) {
                    if (!lastPressed || (lastPressed + interval) < now) {
                        gameAction();
                        this.controls[control].lastPressed = now;
                    }
                }
                else {
                    this.controls[control].lastPressed = null;
                }
            };
        };
        const makeStickChecker = (whichStick) => {
            return () => {
                const now = Date.now();
                const interval = (this.repeatIntervalInSeconds || .25) * 1000;
                const control = this.stickControls[whichStick];
                const { stick, lastPressed } = control;
                const { x: xAxis, y: yAxis } = stick.query();
                const yLean = Math.abs(yAxis);
                const xLean = Math.abs(xAxis);
                if (yLean > xLean) {
                    if (yLean > .5) {
                        if (!lastPressed || (lastPressed + interval) < now) {
                            control.lastPressed = now;
                            this.polledInput = yAxis < 0 ? util_1.INPUT_BUTTON.UP : util_1.INPUT_BUTTON.DOWN;
                        }
                        return;
                    }
                }
                else {
                    if (xLean > .5) {
                        if (!lastPressed || (lastPressed + interval) < now) {
                            control.lastPressed = now;
                            this.polledInput = xAxis < 0 ? util_1.INPUT_BUTTON.LEFT : util_1.INPUT_BUTTON.RIGHT;
                        }
                        return;
                    }
                }
                control.lastPressed = null;
            };
        };
        this.controlCheckers = [
            makeStickChecker('left'),
            makeStickChecker('right'),
            makeChecker('up', () => this.polledInput = util_1.INPUT_BUTTON.UP),
            makeChecker('down', () => this.polledInput = util_1.INPUT_BUTTON.DOWN),
            makeChecker('left', () => this.polledInput = util_1.INPUT_BUTTON.LEFT),
            makeChecker('right', () => this.polledInput = util_1.INPUT_BUTTON.RIGHT),
            makeChecker('action', () => this.polledInput = util_1.INPUT_BUTTON.ACTION),
            makeChecker('undo', () => this.polledInput = util_1.INPUT_BUTTON.UNDO),
            makeChecker('restart', () => this.polledInput = util_1.INPUT_BUTTON.RESTART)
        ];
        // Disable bubbling up events when keys are pressed
        this.boundOnKeyEvents = this.onKeyEvents.bind(this);
        window.addEventListener('keydown', this.boundOnKeyEvents);
        window.addEventListener('keyup', this.boundOnKeyEvents);
        window.addEventListener('keypress', this.boundOnKeyEvents);
    }
    onKeyEvents(evt) {
        if (evt.metaKey || evt.shiftKey || evt.altKey || evt.ctrlKey) {
            return;
        }
        if (window.document.activeElement !== this.table) {
            return;
        }
        if (this.possibleKeys.indexOf(evt.key) >= 0) {
            evt.preventDefault();
        }
    }
    pollControls() {
        if (window.document.activeElement !== this.table) {
            // Ignore if the game is not in focus
            return null;
        }
        this.polledInput = null;
        this.controlCheckers.forEach((fn) => fn());
        const input = this.polledInput;
        this.polledInput = null;
        return input;
    }
    isSomethingPressed() {
        for (const entry of Object.values(this.controls)) {
            if (entry.button.query()) {
                return true;
            }
        }
        return false;
    }
    setKeyRepeatInterval(repeatIntervalInSeconds) {
        this.repeatIntervalInSeconds = repeatIntervalInSeconds;
    }
    dispose() {
        window.removeEventListener('keydown', this.boundOnKeyEvents);
        window.removeEventListener('keyup', this.boundOnKeyEvents);
        window.removeEventListener('keypress', this.boundOnKeyEvents);
    }
}
exports.default = InputWatcher;
//# sourceMappingURL=InputWatcher.js.map